@@ -1,4 +1,3 @@
-# pyre-strict
 """
 WDL data types
 
@@ -27,22 +26,21 @@ coerced to some other desired type, according to the following rules:
 
 1. ``Int`` coerces to ``Float``
 2. ``Boolean``, ``Int``, ``Float``, and ``File`` coerce to ``String``
-3. ``Array[T]`` coerces to ``String`` provided ``T`` does as well
-4. ``T`` coerces to ``T?`` but the reverse is not true in general*
-5. ``Array[T]+`` coerces to ``Array[T]`` but the reverse is not true in general*
-6. ``T`` coerces to ``Array[T]+`` (an array of length 1).
+3. ``String`` coerces to ``File``, ``Int``, and ``Float``
+4. ``Array[T]`` coerces to ``String`` provided ``T`` does as well
+5. ``T`` coerces to ``T?`` but the reverse is not true in general*
+6. ``Array[T]+`` coerces to ``Array[T]`` but the reverse is not true in general*
 
 (*) The reverse coercions are statically permitted in expressions set up with
-``Expr.infer_type(check_quant=False)`` although they may fail at runtime.
+``Expr.infer_type(check_quant=False)`` although they may fail at runtime. This
+also enables coercion of ``T`` to ``Array[T]+`` (an array of length 1).
 
 .. inheritance-diagram:: WDL.Type
    :top-classes: WDL.Type.Base
 """
-from abc import ABC
-from typing import Optional, TypeVar, Tuple
 import copy
-
-TVBase = TypeVar("TVBase", bound="Base")
+from abc import ABC
+from typing import Optional, Tuple, Dict, Iterable, Set
 
 
 class Base(ABC):
@@ -56,23 +54,29 @@ class Base(ABC):
     All instances are immutable.
     """
 
-    _optional: bool  # immutable!!!
+    _optional: bool = False  # immutable!!!
 
-    def coerces(self, rhs: TVBase, check_quant: bool = True) -> bool:
+    # pos is set on Type objects instantiated by the WDL syntax parser (mainly in Decl). Other Type
+    # objects are instantiated in other ways (e.g. Value describing itself), so will not have pos.
+    pos: "Optional[WDL.Error.SourcePosition]" = None
+
+    def coerces(self, rhs: "Base", check_quant: bool = True) -> bool:
         """
         True if this is the same type as, or can be coerced to, ``rhs``.
 
-        :param check_quant: when ``False``, relaxes validation of the optional (?) and nonempty (+) type quantifiers
+        :param check_quant: when ``False``, disables static enforcement of the optional (?) type quantifier
         """
-        if isinstance(rhs, Array) and self.coerces(rhs.item_type, check_quant):
+        if not check_quant and isinstance(rhs, Array) and self.coerces(rhs.item_type, check_quant):
             # coerce T to Array[T]
             return True
-        return (type(rhs).__name__ in [type(self).__name__, "Any"]) and self._check_optional(
-            rhs, check_quant
-        )
+        return (
+            type(self).__name__ == type(rhs).__name__ or isinstance(rhs, Any)
+        ) and self._check_optional(rhs, check_quant)
 
-    def _check_optional(self, rhs: TVBase, check_quant: bool) -> bool:
-        return not (check_quant and (self.optional and not rhs.optional))
+    def _check_optional(self, rhs: "Base", check_quant: bool) -> bool:
+        return not (
+            check_quant and (self.optional and not rhs.optional and not isinstance(rhs, Any))
+        )
 
     @property
     def optional(self) -> bool:
@@ -82,7 +86,17 @@ class Base(ABC):
         True when the type has the optional quantifier, ``T?``"""
         return self._optional
 
-    def copy(self, optional: Optional[bool] = None) -> TVBase:
+    @property
+    def parameters(self) -> Iterable["Base"]:
+        """
+        :type: Iterable[WDL.Type.Base]
+
+        The type's parameters, if any (e.g. item type of Array; left & right
+        types of Pair; etc.)
+        """
+        return []
+
+    def copy(self, optional: Optional[bool] = None) -> "Base":
         """
         copy(self, optional : Optional[bool] = None) -> WDL.Type.Base
 
@@ -97,7 +111,7 @@ class Base(ABC):
     def __str__(self) -> str:
         return type(self).__name__ + ("?" if self.optional else "")
 
-    def __eq__(self, rhs) -> bool:
+    def __eq__(self, rhs: "Base") -> bool:
         return isinstance(rhs, Base) and str(self) == str(rhs)
 
 
@@ -107,10 +121,10 @@ class Any(Base):
     """
 
     def __init__(self, optional: bool = False) -> None:
-        self._optional = optional
+        self._optional = False  # no point, since this unconditionally coerces to anything
 
-    def coerces(self, rhs: TVBase, check_quant: bool = True) -> bool:
-        return self._check_optional(rhs, check_quant)
+    def coerces(self, rhs: Base, check_quant: bool = True) -> bool:
+        return True
 
 
 class Boolean(Base):
@@ -163,7 +177,7 @@ class String(Base):
 
     def coerces(self, rhs: Base, check_quant: bool = True) -> bool:
         ""
-        if isinstance(rhs, File):
+        if isinstance(rhs, (File, Int, Float)):
             return self._check_optional(rhs, check_quant)
         return super().coerces(rhs, check_quant)
 
@@ -207,13 +221,15 @@ class Array(Base):
         """
         return self._nonempty
 
+    @property
+    def parameters(self) -> Iterable[Base]:
+        yield self.item_type
+
     def coerces(self, rhs: Base, check_quant: bool = True) -> bool:
         ""
         if isinstance(rhs, Array):
-            return (
-                self.item_type.coerces(rhs.item_type, check_quant)
-                and (not check_quant or not rhs.nonempty or self.nonempty)
-                and self._check_optional(rhs, check_quant)
+            return self.item_type.coerces(rhs.item_type, check_quant) and self._check_optional(
+                rhs, check_quant
             )
         if isinstance(rhs, String):
             return self.item_type is None or self.item_type.coerces(String())
@@ -222,7 +238,7 @@ class Array(Base):
         return False
 
     def copy(self, optional: Optional[bool] = None, nonempty: Optional[bool] = None) -> Base:
-        ans: Array = super().copy(optional)
+        ans = super().copy(optional)
         if nonempty is not None:
             ans._nonempty = nonempty
         return ans
@@ -240,17 +256,29 @@ class Map(Base):
     The key and value types may be ``Any`` when not known statically, such as in a literal empty map ``{}``.
     """
 
-    def __init__(self, item_type: Tuple[Base, Base], optional: bool = False) -> None:
+    literal_keys: Optional[Set[str]]
+    ""
+    # Special use: Map[String,_] literal stores the key names here for potential use in
+    # struct coercions where we need them. (Normally the Map type would record the common
+    # type of the keys but not the keys themselves.)
+
+    def __init__(
+        self,
+        item_type: Tuple[Base, Base],
+        optional: bool = False,
+        literal_keys: Optional[Set[str]] = None,
+    ) -> None:
         self._optional = optional
         if item_type is None:
             item_type = (Any(), Any())
         self.item_type = item_type
+        self.literal_keys = literal_keys
 
     def __str__(self) -> str:
         return (
             "Map["
             + (
-                str(self.item_type[0]) + "," + str(self.item_type[1])  # pyre-fixme
+                str(self.item_type[0]) + "," + str(self.item_type[1])
                 if self.item_type is not None
                 else ""
             )
@@ -258,15 +286,34 @@ class Map(Base):
             + ("?" if self.optional else "")
         )
 
+    @property
+    def parameters(self) -> Iterable[Base]:
+        yield self.item_type[0]
+        yield self.item_type[1]
+
     def coerces(self, rhs: Base, check_quant: bool = True) -> bool:
         ""
         if isinstance(rhs, Map):
-            # pyre-fixme
             return (
                 self.item_type[0].coerces(rhs.item_type[0], check_quant)
                 and self.item_type[1].coerces(rhs.item_type[1], check_quant)
                 and self._check_optional(rhs, check_quant)
             )
+        if isinstance(rhs, StructInstance) and self.literal_keys is not None:
+            # struct assignment from map literal: the map literal must contain all non-optional
+            # struct members, and the value type must be coercible to those member types
+            rhs_members = rhs.members
+            assert rhs_members is not None
+            rhs_keys = set(rhs_members.keys())
+            if self.literal_keys - rhs_keys:
+                return False
+            for k in self.literal_keys:
+                if not self.item_type[1].coerces(rhs_members[k], check_quant):
+                    return False
+            for opt_k in rhs_keys - self.literal_keys:
+                if not rhs_members[opt_k].optional:
+                    return False
+            return True
         if isinstance(rhs, Any):
             return self._check_optional(rhs, check_quant)
         return False
@@ -299,6 +346,11 @@ class Pair(Base):
             + ("?" if self.optional else "")
         )
 
+    @property
+    def parameters(self) -> Iterable[Base]:
+        yield self.left_type
+        yield self.right_type
+
     def coerces(self, rhs: Base, check_quant: bool = True) -> bool:
         ""
         if isinstance(rhs, Pair):
@@ -310,3 +362,122 @@ class Pair(Base):
         if isinstance(rhs, Any):
             return self._check_optional(rhs, check_quant)
         return False
+
+
+class StructInstance(Base):
+    """
+    Type of an instance of a struct
+
+    Not to be confused with struct type definition, :class:`WDL.Tree.StructTypeDef`. To find the
+    ``WDL.Tree.StructTypeDef`` in the current ``doc: WDL.Tree.Document`` corresponding to
+    ``ty: WDL.Type.StructInstance``, use ``doc.struct_typedefs[ty.type_name]``.
+    """
+
+    type_name: str
+    """
+    :type: str
+
+    The struct type name with which the instance is declared; note that the
+    same struct type can go by different names.
+    """
+
+    members: Optional[Dict[str, Base]]
+    """
+    :type: Dict[str,WDL.Type.Base]
+
+    Names and types of the struct members, from the struct type definition
+    (available after typechecking)
+    """
+
+    def __init__(self, type_name: str, optional: bool = False) -> None:
+        self._optional = optional
+        self.type_name = type_name
+        self.members = None
+
+    def __str__(self) -> str:
+        return self.type_name + ("?" if self.optional else "")
+
+    def coerces(self, rhs: Base, check_quant: bool = True) -> bool:
+        ""
+        if isinstance(rhs, StructInstance):
+            return self.type_id == rhs.type_id and self._check_optional(rhs, check_quant)
+        if isinstance(rhs, Any):
+            return self._check_optional(rhs, check_quant)
+        return False
+
+    @property
+    def type_id(self) -> str:
+        """
+        :type: str
+
+        A string uniquely describing the member names and types, excluding the struct type name; useful to identify aliased struct types.
+        """
+        assert isinstance(self.members, dict)
+        return _struct_type_id(self.members)
+
+    @property
+    def parameters(self) -> Iterable[Base]:
+        assert self.members is not None
+        return self.members.values()
+
+
+def _struct_type_id(members: Dict[str, Base]) -> str:
+    # generates a content hash of the struct type definition, used to recognize
+    # equivalent struct types going by different aliases
+    ans = []
+    for (name, ty) in sorted(members.items()):
+        if isinstance(ty, StructInstance):
+            assert ty.members
+            ty = _struct_type_id(ty.members) + ("?" if ty.optional else "")
+        else:
+            ty = str(ty)
+        ans.append(name + " : " + ty)
+    return "struct(" + ", ".join(ans) + ")"
+
+
+class Object(Base):
+    ""
+    # In WDL 1.0, struct instances are created by coercion from object
+    # literals. So we need something to represent the type of an object literal
+    # (a bag of keys and values) prior to its coercion to a named struct type.
+    # But we hide this from docs to avoid confusion with general Object
+    # support.
+
+    members: Dict[str, Base]
+
+    def __init__(self, members: Dict[str, Base]) -> None:
+        self.members = members
+
+    def __str__(self) -> str:
+        ans = []
+        for name, ty in sorted(self.members.items()):
+            ans.append(name + " : " + str(ty))
+        return "object(" + ", ".join(ans) + ")"
+
+    @property
+    def parameters(self) -> Iterable[Base]:
+        return self.members.values()
+
+    def coerces(self, rhs: Base, check_quant: bool = True) -> bool:
+        if isinstance(rhs, (StructInstance, Object)):
+            rhs_members = rhs.members
+            assert rhs_members is not None
+            # Check whether our keys match the struct members, and our types
+            # are coercible to the respective member types.
+            # TODO: in the event of StaticTypeMismatch errors, this may produce
+            # unwieldy error messages
+            self_keys = set(self.members.keys())
+            rhs_keys = set(rhs_members.keys())
+            if self_keys - rhs_keys:
+                return False
+            for k in self_keys:
+                if not self.members[k].coerces(rhs_members[k], check_quant):
+                    return False
+            for opt_k in rhs_keys - self_keys:
+                # object literal may omit optional struct fields
+                if not rhs_members[opt_k].optional:
+                    return False
+            return True
+        if isinstance(rhs, Any):
+            return self._check_optional(rhs, check_quant)
+        return False
